/***
 * CopyPolicy: GNU Public License 2 applies
 * Copyright (C) 1994-1997 Heiko Eissfeldt heiko@colossus.escape.de
 *
 * Interface module for cdrom drive access
 *
 * Two interfaces are possible.
 *
 * 1. using 'cooked' ioctls()
 *    disadvantages: available for atapi, sbpcd and cdu31a drives only.
 *
 * 2. using the generic scsi device (for details see SCSI Prog. HOWTO).
 *    NOTE: a bug/misfeature in the kernel requires blocking signal
 *          SIGINT during SCSI command handling. Once this flaw has
 *          been removed, the sigprocmask SIG_BLOCK and SIG_UNBLOCK calls
 *          should removed, thus saving context switches.
 *
 * For testing purposes I have added a third simulation interface.
 *
 * Version 0.8: used experiences of Jochen Karrer.
 *              SparcLinux port fixes
 *              AlphaLinux port fixes
 *
 */
#if 0
#define SIM_CD
#endif

#include <stdio.h>
#include <stdlib.h>
#if defined (HAVE_UNISTD_H) && (HAVE_UNISTD_H == 1)
#include <sys/types.h>
#include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <assert.h>

#include <sys/ioctl.h>
#include <sys/stat.h>


#if defined(_AIX)
# include <sys/scsi.h>
# include <sys/scsi_buf.h>
# include <sys/scdisk.h>
#  define SG_BIG_BUFF 4096	/* FIXME: should be the size of page */
#else
#if defined(HPUX)
# include <sys/scsi.h>
#  define SG_BIG_BUFF 4096	/* FIXME: should be the size of page */
struct sg_header {};
#else

#include "lowlevel.h"
/* some include file locations have changed with newer kernels */
#if defined (__linux__)
# if LINUX_VERSION_CODE > 0x10300 + 97
#  include <scsi/sg.h>
#  include <scsi/scsi.h>
# else /* old stuff */
#  include <linux/../../drivers/scsi/sg.h>
#  include <linux/../../drivers/scsi/scsi.h>
# endif

/* Hack hack hack hack hack... makes the code clean, though. Zygo was here */
# ifndef SG_BIG_BUFF
#  define SG_BIG_BUFF 4096	/* FIXME: should be the size of page */
# endif
/* HACK! the kernel header files were misdesigned for
 *  user applications.
 * #include <linux/blkdev.h> did not work
 */
struct request {		/* this is and may not used */
  int dummy;
};
#endif
#endif /* defined HPUX */
#endif /* defined _AIX */

#include "byteorder.h"
#include "mycdrom.h"
#include "cdda2wav.h"
#include "interface.h"
#include "semshm.h"
#include "setuid.h"
#include "ringbuff.h"
#include "global.h"
#include "paranoia.h"
#include "scsi_cmds.h"

#include <scsitransp.h>
sigset_t signalset;    /* to circumvent a kernel problem with the generic driver */
struct sigaction sigac;
unsigned interface;

int trackindex_disp = 0;

TOC g_toc [MAXTRK]; /* 100 */
unsigned char MCN[15];

void     (*EnableCdda) (int Switch);
unsigned (*ReadToc) ( TOC *ptoc );
void     (*ReadCdRom) (unsigned char *p, long lSector, unsigned long SectorBurstVal );
subq_chnl *(*ReadSubQ) ( unsigned char format, unsigned char track );
void     (*SelectSpeed) ( unsigned speed );

/* buffer for cdrom audio data sector */
unsigned char *bufferCdda;

#if 0
/* process a complete scsi cmd. Use either generic or ioctl interface. */
static int handle_scsi_cmd(unsigned cmd_len, 
			   unsigned in_size, 
			   unsigned char *i_buff, 
			   unsigned out_size, 
			   unsigned char *o_buff,
			   int expect_result)
{
#if defined(HPUX)
    struct sctl_io sctl_io;

    if (!cmd_len) return -1;
    if (!i_buff) return -1;
	
/* fprintf(stderr, "doing scsi: cmd=%x\n",*i_buff); */
	
    if (in_size && out_size) {
      /* cannot handle reading & writing to the device with one call */
      return -1;
    }
    
    memset(&sctl_io, 0, sizeof sctl_io);   /* clear reserved fields */
	
    sctl_io.cdb_length = cmd_len;
    memcpy(&(sctl_io.cdb[0]), i_buff, cmd_len);      /* copy command */
	
    if (in_size) {
      /* we'll write some data to the SCSI device */
      sctl_io.flags = 0;
      sctl_io.data_length = in_size;
      sctl_io.data = i_buff + cmd_len;
    } else {
      /* we'll read some data from the SCSI device */
      sctl_io.flags = SCTL_READ;
      sctl_io.data_length = out_size;
      sctl_io.data = (out_size)? o_buff : 0;		
    }
	
    sctl_io.max_msecs = 10000;              /* allow 10 seconds for cmd */
    
    if (ioctl(global.cd_fd, SIOC_IO, &sctl_io) < 0) {
      perror("scsi_io");
      exit(-1);
    }
    
    if (sctl_io.cdb_status == S_GOOD) {
      return 0; /* everything is fine... */
    }
    
    if (sctl_io.cdb_status == S_BUSY) {
      fprintf(stderr, "sctl_io: device is busy\n");
      return -1;
    }
    
    if (sctl_io.cdb_status == S_CHECK_CONDITION) {
      fprintf(stderr, "sctl_io: S_CHECK_CONDITION\n");
      if (sctl_io.sense_status != S_GOOD) {
	fprintf(stderr, "sctl_io: sense_status != S_GOOD\n");
      } else {
	fprintf(stderr, "sctl_io: sense_xfer=%d, sense_data=%x %x %x %x %x %x ...\n",
		sctl_io.sense_xfer, sctl_io.sense[0], sctl_io.sense[1], sctl_io.sense[2],
		sctl_io.sense[3], sctl_io.sense[4], sctl_io.sense[5]);			
      }
      return -1;
    }
	
    return 0;	

#else
#if defined(_AIX)
    struct scsi_iocmd scsi_io;

    if (!cmd_len) return -1;
    if (!i_buff) return -1;
	
/* fprintf(stderr, "doing scsi: cmd=%x\n",*i_buff); */
	
    if (in_size && out_size) {
      /* cannot handle reading & writing to the device with one call */
      return -1;
    }
    
    memset(&scsi_io, 0, sizeof scsi_io);   /* clear reserved fields */
    memcpy(&(scsi_io.scsi_cdb), i_buff, cmd_len);      /* copy command */
	
    scsi_io.command_length = cmd_len;
#ifdef NOTYET
    scsi_io.lun_id = lun;
    scsi_io.scsi_id = id;
    scsi_io.scsi_cdb[1] |= lun;
#endif

    if (in_size) {
      /* we'll write some data to the SCSI device */
      scsi_io.flags = B_WRITE;
      scsi_io.data_length = in_size;
      scsi_io.buffer = i_buff + cmd_len;
    } else {
      /* we'll read some data from the SCSI device */
      scsi_io.flags = B_READ;
      scsi_io.data_length = out_size;
      scsi_io.buffer = (out_size)? o_buff : 0;		
    }
	
    scsi_io.timeout_value = 10;              /* allow 10 seconds for cmd */

    if (ioctl(global.cd_fd, DKIOCMD, &scsi_io) < 0) {
      perror("scsi_io");
      exit(-1);
    }
    
    if ((scsi_io.status_validity & (SC_SCSI_ERROR | SC_ADAPTER_ERROR)) == 0) { 
      return 0; /* everything is fine... */
    }
    
    if (scsi_io.status_validity & SC_SCSI_ERROR) { 
      fprintf(stderr, "SCSI Error: status 0x%02x\n", scsi_io.scsi_bus_status);
    }
    if (scsi_io.status_validity & SC_ADAPTER_ERROR) { 
      fprintf(stderr, "SCSI Adapter Error: status 0x%02x\n", scsi_io.adapter_status);
    }
    return -1;
    
#else
#if defined (__linux__)
    int status = 0;
    unsigned char *buff_p;
    struct sg_header *sg_hd;

    /* safety checks */
    if (!cmd_len) return -1;
    if (!i_buff) return -1;
    if (OFF + cmd_len + in_size > SG_BIG_BUFF) return -1;
    if (OFF + out_size > SG_BIG_BUFF) return -1;

    if (!o_buff) out_size = 0;

    if (out_size > in_size + cmd_len) {
	buff_p = o_buff;
	memcpy(o_buff + OFF, i_buff + OFF, in_size + cmd_len);
    } else {
	buff_p = i_buff;
    }

    if (interface == GENERIC_SCSI) {
	/* generic scsi device services */
	sg_hd = (struct sg_header *) buff_p;
	sg_hd->reply_len = OFF + out_size;
	sg_hd->twelve_byte = cmd_len == 12;
	sg_hd->result = 0;	/* is not cleared by the write call */
#if	0
	sg_hd->pack_len; /* not used */
	sg_hd->pack_id;	/* not used */
	sg_hd->other_flags; 	/* not used */
#endif
        sigprocmask ( SIG_BLOCK, &signalset, NULL );
	status = write( global.cd_fd, buff_p, OFF + cmd_len + in_size );
	if ( status < 0 || (unsigned) status != OFF + cmd_len + in_size || sg_hd->result ) {
	    if (status == -1) {
		perror("generic scsi write: ");
		exit(status);
	    } else {
	       fprintf( stderr, "\ngeneric scsi write status = 0x%x, result = 0x%x cmd = %x\n", 
		       status, sg_hd->result, i_buff[OFF] );
	    }
	    return status;
	}

	status = read( global.cd_fd, buff_p, OFF + out_size);
	if ( status < 0 || (unsigned) status != OFF + out_size || (sg_hd->result && !expect_result) ) {
	  if (sg_hd->sense_buffer[0]==0xf0 && /* hack to make Plextor 6plex happy */
	      sg_hd->sense_buffer[1]==0x0 &&
	      sg_hd->sense_buffer[2]==0x5 &&
	      /* 3-6 are some sort of block address */
	      sg_hd->sense_buffer[7]==0xa &&
	      sg_hd->sense_buffer[8]==0x0 &&
	      sg_hd->sense_buffer[9]==0x0 &&
	      sg_hd->sense_buffer[10]==0x0 &&
	      sg_hd->sense_buffer[11]==0x0 &&
	      sg_hd->sense_buffer[12]==0xbf &&
	      sg_hd->sense_buffer[13]==0x0 &&
	      sg_hd->sense_buffer[14]==0x0 &&
	      sg_hd->sense_buffer[15]==0x0) {
	    ;
	  } else {
	    fprintf( stderr, "\nread(generic) status = 0x%x, result = 0x%x cmd = %x\n", 
		    status, sg_hd->result, i_buff[OFF] );
	    fprintf( stderr, "read(generic) sense "
		  "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x\n", 
		  sg_hd->sense_buffer[0],	     sg_hd->sense_buffer[1],
		  sg_hd->sense_buffer[2],	     sg_hd->sense_buffer[3],
		  sg_hd->sense_buffer[4],	     sg_hd->sense_buffer[5],
		  sg_hd->sense_buffer[6],	     sg_hd->sense_buffer[7],
		  sg_hd->sense_buffer[8],	     sg_hd->sense_buffer[9],
		  sg_hd->sense_buffer[10],	     sg_hd->sense_buffer[11],
		  sg_hd->sense_buffer[12],	     sg_hd->sense_buffer[13],
		  sg_hd->sense_buffer[14],	     sg_hd->sense_buffer[15]
		  );
	  }
	  sigprocmask ( SIG_UNBLOCK, &signalset, NULL );
	  return 1;
	} else
	    /* get output back */
	    if (out_size && out_size <= in_size + cmd_len)
		memcpy(o_buff, i_buff, out_size + OFF);
	/* Look if we got what we expected to get */
	sigprocmask ( SIG_UNBLOCK, &signalset, NULL );
	if ((unsigned) status == OFF + out_size) status = 0; /* got them all */
    }
    return status;
#endif
#endif
#endif
  return 0;
}
#endif

/* hook */
static void Dummy ( void )
{
}

#if !defined(SIM_CD)
/* set the correct command set for *different*
 * vendor specific implementations. Was this really necessary, folks?
 */
static void SetupSCSI( void )
{
    unsigned char *p;

    if (interface != GENERIC_SCSI) {
	/* unfortunately we have the wrong interface and are
	 * not able to change on the fly */
	fprintf(stderr, "The generic SCSI interface and devices are required\n");
	exit(1);
    }

    /* check for the correct type of unit. */
    p = Inquiry();
#if defined (__linux__)
    if (*p != TYPE_ROM && *p != TYPE_WORM) {
#else
#ifndef SCSI_CDROM
#define SCSI_CDROM 5
#define SCSI_WORM  4
#endif
    if (*p != SCSI_CDROM && *p != SCSI_WORM) {
#endif
	fprintf(stderr, "this is neither a scsi cdrom nor a worm device\n");
	exit(1);
    }

    /* generic Sony type defaults */
    density = 0x0;
    EnableCdda = (void (*)(int))Dummy;
    ReadCdRom = ReadCdda12;
    SelectSpeed = SpeedSelectSCSISony;

    /* check for brands and adjust special peculiaritites */

    /* If your drive is not treated correctly, you can adjust some things
       here:

       global.littleendian: should be to 1, if the CDROM drive or CD-Writer
		  delivers the samples in the native byteorder of the audio cd
		  (LSB first).
		  HP CD-Writers need it set to 0.
       NOTE: If you get correct wav files when using sox with the '-x' option,
             the endianess is wrong. You can use the -C option to specify
	     the value of global.littleendian.

     */

    allow_atapi(1);

    switch (get_mmc()) {
     case 2:      /* SCSI-3 cdrom drive with accurate audio stream */
       if ( global.littleendian == -1 )
	 global.littleendian = 1;
       global.overlap = 0;
       ReadCdRom = ReadCddaMMC12;
       SelectSpeed = SpeedSelectSCSIMMC;
       break;
     case 1:      /* SCSI-3 cdrom drive with no accurate audio stream */
       if ( global.littleendian == -1 )
	 global.littleendian = 1;
       ReadCdRom = ReadCddaMMC12;
       SelectSpeed = SpeedSelectSCSIMMC;
       break;
     case 0:      /* non SCSI-3 cdrom drive */
    if (!memcmp(p+8,"TOSHIBA", 7) ||
        !memcmp(p+8,"IBM", 3) ||
        !memcmp(p+8,"DEC", 3)) {
	density = 0x82;
	EnableCdda = EnableCddaModeSelect;
 	ReadCdRom = ReadStandard;
        SelectSpeed = SpeedSelectSCSIToshiba;
	if ( global.littleendian == -1 )
		global.littleendian = 1;
    } else if (!memcmp(p+8,"IMS",3) ||
               !memcmp(p+8,"KODAK",5) ||
               !memcmp(p+8,"RICOH",5) ||
               !memcmp(p+8,"HP",2) ||
               !memcmp(p+8,"PHILIPS",7) ||
               !memcmp(p+8,"PLASMON",7) ||
               !memcmp(p+8,"GRUNDIG CDR100IPW",17) ||
               !memcmp(p+8,"MITSUMI CD-R ",13)) {
	EnableCdda = EnableCddaModeSelect;
	ReadCdRom = ReadStandard;
        SelectSpeed = SpeedSelectSCSIPhilipsCDD2600;

	/* treat all of these as bigendian */
	if ( global.littleendian == -1 )
		global.littleendian = 0;

	/* no overlap reading for cd-writers */
	global.overlap = 0;
    } else if (!memcmp(p+8,"YAMAHA",6)) {
	EnableCdda = EnableCddaModeSelect;
        SelectSpeed = SpeedSelectSCSIYamaha;

	/* no overlap reading for cd-writers */
	global.overlap = 0;
	if ( global.littleendian == -1 )
		global.littleendian = 1;
    } else if (!memcmp(p+8,"PLEXTOR",7) ||
               !memcmp(p+8,"SONY",4)) {
/*	global.overlap = 0; */
	if ( global.littleendian == -1 )
		global.littleendian = 1;
        if (!memcmp(p+16, "CD-ROM CDU55E",13)) {
	   ReadCdRom = ReadCddaMMC12;
	}
    } else if (!memcmp(p+8,"NEC",3)) {
	ReadCdRom = ReadCdda10;
        SelectSpeed = SpeedSelectSCSINEC;
	if ( global.littleendian == -1 )
		global.littleendian = 1;
        if (!memcmp(p+29,"5022.0r",3)) /* I assume all versions of the 502 require this? */
               global.overlap = 0;           /* no overlap reading for NEC CD-ROM 502 */
    }
    } /* switch (get_mmc) */

    ReadToc = (!memcmp(p+8, "IMS", 3)) ? ReadTocCdrSCSI : ReadTocSCSI;
    ReadSubQ = ReadSubQSCSI;

    /* look if caddy is loaded */
#if defined(HPUX)
    {
      /* need to test for medium a few times to get a reliable answer... */
      int i;
      for (i=0; i < 3; i++) {
	if (TestForMedium()) return;
      }
    }
#endif
    while (!TestForMedium()) {
	fprintf(stderr,"load cdrom please and press enter");
	getchar();
    }
}

#if defined (__linux__)
/*******************************************************************
 *
 *	cooked ioctl section
 *
 */
static struct cdrom_tochdr hdr;
static struct cdrom_tocentry entry[100];
static struct cdrom_read_audio arg;
static int err;

/* read the table of contents (toc) via the ioctl interface */
static unsigned ReadToc_cooked ( TOC *toc )
{
    unsigned i;
    unsigned tracks;

    /* get TocHeader to find out how many entries there are */
    err = ioctl( global.cd_fd, CDROMREADTOCHDR, &hdr );
    if ( err != 0 ) {
	/* error handling */
	if (err == -1) {
	    if (errno == EPERM)
		fprintf( stderr, "Please run this program setuid root.\n");
	    perror("cooked: Read TOC ");
	    exit( err );
	} else {
	    fprintf( stderr, "can't get TocHeader (error %d).\n", err );
	    exit( -1 );
	}
    }
    /* get all TocEntries */
    for ( i = 0; i < hdr.cdth_trk1; i++ ) {
	entry[i].cdte_track = 1+i;
	entry[i].cdte_format = CDROM_LBA;
	err = ioctl( global.cd_fd, CDROMREADTOCENTRY, &entry[i] );
	if ( err != 0 ) {
	    /* error handling */
	    fprintf( stderr, "can't get TocEntry #%d (error %d).\n", i+1, err );
	    exit( -1 );
	}
    }
    entry[i].cdte_track = CDROM_LEADOUT;
    entry[i].cdte_format = CDROM_LBA;
    err = ioctl( global.cd_fd, CDROMREADTOCENTRY, &entry[i] );
    if ( err != 0 ) {
	/* error handling */
	fprintf( stderr, "can't get TocEntry LEADOUT (error %d).\n", err );
	exit( -1 );
    }
    tracks = hdr.cdth_trk1+1;
    for (i = 0; i < tracks; i++) {
        toc[i].bFlags = (entry[i].cdte_adr << 4) | (entry[i].cdte_ctrl & 0x0f);
        toc[i].bTrack = entry[i].cdte_track;
        toc[i].dwStartSector = entry[i].cdte_addr.lba;
    }
    return --tracks;           /* without lead-out */
}

/* read 'SectorBurst' adjacent sectors of audio sectors 
 * to Buffer '*p' beginning at sector 'lSector'
 */
static void ReadCdRom_cooked (unsigned char *p, long lSector, unsigned long SectorBurstVal )
{
  int retry_count=0;
  static int nothing_read = 1;

/* read 2352 bytes audio data */
  arg.addr.lba = lSector;
  arg.addr_format = CDROM_LBA;
  arg.nframes = SectorBurstVal;
  arg.buf = &p[0];

  do {
    err = ioctl(global.cd_fd, CDROMREADAUDIO, &arg);
    retry_count++;
  } while ((err) && (retry_count < 30));
  if (err != 0) {
      /* error handling */
      if (err == -1) {
	  if (nothing_read && (errno == EINVAL || errno == EIO))
	      fprintf( stderr, "Sorry, this driver and/or drive does not support cdda reading.\n");
	  perror("cooked: Read cdda ");
          fprintf(stderr, " sector %ld, buffer %p\n", lSector, p);
      } else {
	  fprintf(stderr, "can't read frame #%d (error %d).\n", 
		  arg.addr.lba, err);
      }
  } else {
    nothing_read = 0;
  }
}

/* request sub-q-channel information. This function may cause confusion
 * for a drive, when called in the sampling process.
 */
static subq_chnl *ReadSubQ_cooked ( unsigned char format, unsigned char track )
{
    struct cdrom_subchnl sub_ch;

    switch (format) {
      case GET_CATALOGNUMBER:
      if (!(err = ioctl(global.cd_fd, CDROM_GET_MCN, (struct cdrom_mcn *) SubQbuffer))) {
          subq_chnl *SQp = (subq_chnl *) SubQbuffer;
	  subq_catalog *SQPp = (subq_catalog *) &SQp->data;

          memmove(SQPp->media_catalog_number, SQp, 14);
          SQPp->mc_valid = 0x80;
      } else {
          return NULL;
      }
      break;
      case GET_POSITIONDATA:
      if (!(err = ioctl(global.cd_fd, CDROMSUBCHNL, &sub_ch))) {
	  /* copy to SubQbuffer */
	  subq_chnl *SQp = (subq_chnl *) (SubQbuffer);
	  subq_position *SQPp = (subq_position *) &SQp->data;
	  SQp->audio_status 	= sub_ch.cdsc_audiostatus;
	  SQp->format 		= sub_ch.cdsc_format;
	  SQp->control_adr	= (sub_ch.cdsc_adr << 4) | (sub_ch.cdsc_ctrl & 0x0f);
	  SQp->track 		= sub_ch.cdsc_trk;
	  SQp->index 		= sub_ch.cdsc_ind;
	  SQPp->abs_min 	= sub_ch.cdsc_absaddr.msf.minute;
	  SQPp->abs_sec 	= sub_ch.cdsc_absaddr.msf.second;
	  SQPp->abs_frame 	= sub_ch.cdsc_absaddr.msf.frame;
	  SQPp->trel_min 	= sub_ch.cdsc_reladdr.msf.minute;
	  SQPp->trel_sec 	= sub_ch.cdsc_reladdr.msf.second;
	  SQPp->trel_frame 	= sub_ch.cdsc_reladdr.msf.frame;
      } else {
	  if (err == -1) {
	      if (errno == EPERM)
		  fprintf( stderr, "Please run this program setuid root.\n");
	      perror("cooked: Read subq ");
	      exit( err );
	  } else {
	      fprintf(stderr, "can't read sub q channel (error %d).\n", err);
	      exit( -1 );
	  }
      }
      break;
      default:
          return NULL;
    } /* switch */
  return (subq_chnl *)(SubQbuffer);
}

/* Speed control */
static void SpeedSelect_cooked( unsigned speed )
{
#ifdef CDROM_SELECT_SPEED
    if (!(err = ioctl(global.cd_fd, CDROM_SELECT_SPEED, &speed))) {
    } else {
	if (err == -1) {
	    if (errno == EPERM)
		fprintf( stderr, "Please run this program setuid root.\n");
	    perror("cooked: Speed select ");
	    /*exit( err ); */
	} else {
	    fprintf(stderr, "can't set speed %d (error %d).\n", speed, err);
	    exit( -1 );
	}
    }
#endif
}

/* set function pointers to use the ioctl routines */
static void SetupCookedIoctl( char *pdev_name )
{
#if (HAVE_ST_RDEV == 1)
    struct stat statstruct;

    if (fstat(global.cd_fd, &statstruct)) {
      fprintf(stderr, "cannot stat cd %d (%s)\n",global.cd_fd, pdev_name);
      exit(1);
    }
    switch ((int)(statstruct.st_rdev >> 8L)) {
    case CDU31A_CDROM_MAJOR:	/* sony cdu-31a/33a */
        if (global.nsectors >= 14) {
	  global.overlap = 10;
	}
        break;
    case MATSUSHITA_CDROM_MAJOR:	/* sbpcd 1 */
    case MATSUSHITA_CDROM2_MAJOR:	/* sbpcd 2 */
    case MATSUSHITA_CDROM3_MAJOR:	/* sbpcd 3 */
    case MATSUSHITA_CDROM4_MAJOR:	/* sbpcd 4 */
        /* some are more compatible than others */
	err = ioctl(global.cd_fd, CDROMAUDIOBUFSIZ, global.nsectors);
        if (err == -1) {
	  perror("ioctl(CDROMAUDIOBUFSIZ)");
        }
	break;
    }
#endif
    EnableCdda = (void (*)(int))Dummy;
    ReadCdRom = ReadCdRom_cooked;
    ReadToc = ReadToc_cooked;
    ReadSubQ = ReadSubQ_cooked;
    SelectSpeed = SpeedSelect_cooked;
}
#endif

/********************** General setup *******************************/

/* As the name implies, interfaces and devices are checked.  We also
   adjust nsectors, overlap, and interface for the first time here.
   Any unnecessary privileges (setuid, setgid) are also dropped here.
*/
static void Check_interface_for_device( struct stat *statstruct, char *pdev_name)
{
    static const char *int_names[]={
       "generic_scsi", "cooked_ioctl"
    };


#if !defined (STAT_MACROS_BROKEN) || (STAT_MACROS_BROKEN != 1)
    if (!S_ISCHR(statstruct->st_mode) &&
	!S_ISBLK(statstruct->st_mode)) {
      fprintf(stderr, "%s is not a device\n",pdev_name);
      exit(1);
    }
#endif

#if defined (HAVE_ST_RDEV) && (HAVE_ST_RDEV == 1)
    switch ((int) (statstruct->st_rdev >> 8L)) {
#if defined (__linux__)
    case SCSI_GENERIC_MAJOR:	/* generic */
#else
    default:			/* ??? what is the proper value here */
#endif
#if !defined (STAT_MACROS_BROKEN) || (STAT_MACROS_BROKEN != 1)
       if (!S_ISCHR(statstruct->st_mode)) {
	 fprintf(stderr, "%s is not a char device\n",pdev_name);
	 exit(1);
       }
       if (interface != GENERIC_SCSI) {
	 if (!S_ISCHR(statstruct->st_mode)) {
	   fprintf(stderr, "%s is not a char device\n",pdev_name);
	   exit(1);
	 }
	    
	 fprintf(stderr, "wrong interface (%s) for this device (%s)\n"
                 "set to generic_scsi\n",int_names[interface], pdev_name);
	 interface = GENERIC_SCSI;
       }
#endif
       global.nsectors = scsi_bufsize(100*1024*1024)/CD_FRAMESIZE_RAW;
       break;
#if defined (__linux__)
    case SCSI_CDROM_MAJOR:     /* scsi cd */
    default:
	if (!S_ISBLK(statstruct->st_mode)) {
	    fprintf(stderr, "%s is not a block device\n",pdev_name);
	    exit(1);
	}
	if (interface != COOKED_IOCTL) {
	    fprintf(stderr, "cdrom device (%s). "
		    "Setting interface to cooked_ioctl.\n", pdev_name);
	    interface = COOKED_IOCTL;
	}
	break;
#endif
    }
#endif
    if (global.overlap >= global.nsectors)
      global.overlap = global.nsectors-1;
}

/* open the scsi device */
static int open_scsi(char *scsidev, int timeout, int be_verbose)
{
	int	x1, x2, x3;
	int	n = 0;

	if (timeout >= 0)
		deftimeout = timeout;

	if (scsidev != NULL)
		n = sscanf(scsidev, "%d,%d,%d", &x1, &x2, &x3);
	if (n == 3) {
		scsibus = x1;
		target = x2;
		lun = x3;
	} else if (n == 2) {
		scsibus = 0;
		target = x1;
		lun = x2;
#if 0
	} else if (scsidev != NULL) {
		printf("WARNING: device not valid, trying to use default target...\n");
		scsibus = 0;
		target = 6;
		lun = 0;
#endif
	}
	if (be_verbose && scsidev != NULL) {
		printf("scsidev: '%s'\n", scsidev);
		printf("scsibus: %d target: %d lun: %d\n",
						scsibus, target, lun);
	}
	global.target = target;
	global.lun = lun;

	return (scsi_open());
}


/* open the cdrom device */
static int OpenCdRom ( char *pdev_name )
{
  int cd_filedesc;
  int have_named_device = 0;
  struct stat statstruct, fstatstruct;

  /*  The device (given by pdevname) can be:
      a. an SCSI device specified with a /dev/xxx name,
      b. an SCSI device specified with bus,target,lun numbers,
      c. a non-SCSI device such as ATAPI or proprietary CDROM devices.
   */

  if (memcmp(pdev_name, "/dev/", 5) != 0) {
      /* map bus, target id, lun to corresponding device */
  } else {
    have_named_device = 1;
  }

  if (have_named_device) {
    if (stat(pdev_name, &statstruct)) {
      fprintf(stderr, "cannot stat device %s\n", pdev_name);
      exit(1);
    } else {
      Check_interface_for_device( &statstruct, pdev_name );
    }
  }

  if (interface == GENERIC_SCSI) {
#ifdef oldstyle_interface
#if defined(HPUX)
      /* need root privileges */
      needroot(1);
      cd_filedesc = open(pdev_name,O_RDWR);
      dontneedroot();
#else
#if defined(_AIX)
      /* need root privileges */
      needroot(1);
      cd_filedesc = open(pdev_name,O_RDWR);
      dontneedroot();
#else
#if defined(__linux__)
      /* may need special group privileges if there is a group for SCSI
         generic devices and we are setgid. */
      needgroup(0);
      cd_filedesc = open(pdev_name,O_RDWR);
      dontneedgroup();
#endif
#endif
#endif
#else
      needgroup(0);
      cd_filedesc = open_scsi(pdev_name, 100, 0);
      dontneedgroup();
#endif
  } else {
      needgroup(0);
      cd_filedesc = open(pdev_name,O_RDONLY);
      dontneedgroup();
  }


  if (cd_filedesc < 0) {
    fprintf(stderr, "open(%s) in file %s, line %d: ",pdev_name, __FILE__, __LINE__);
    perror("");
    fprintf(stderr, "Most probably you did not define your SCSI device.\n");
    fprintf(stderr, "Set the CDR_DEVICE environment variable or use the -D option.\n");
    fprintf(stderr, "You can also define the default device in the Makefile.\n");
    exit(1);
  }

  /* Do final security checks here */
  if (fstat(cd_filedesc, &fstatstruct)) {
    fprintf(stderr, "Could not fstat %s (fd %d): ", pdev_name, cd_filedesc);
    perror("");
    exit(1);
  }
  Check_interface_for_device( &fstatstruct, pdev_name );

  /* Watch for race conditions */
  if (have_named_device 
     && (fstatstruct.st_dev != statstruct.st_dev ||
         fstatstruct.st_ino != statstruct.st_ino)) {
    fprintf(stderr,"Race condition attempted in OpenCdRom.  Exiting now.\n");
    exit(1);
  }
  return cd_filedesc;
}
#endif /* SIM_CD */

/******************* Simulation interface *****************/
#if	defined SIM_CD
static unsigned long sim_pos=0;

/* read 'SectorBurst' adjacent sectors of audio sectors 
 * to Buffer '*p' beginning at sector 'lSector'
 */
static void ReadCdRom_sim (unsigned char *p, long lSector, unsigned long SectorBurstVal )
{
  unsigned int loop=0;
  short *q = (short *) p;
  int joffset = 0;

#if 0
  /* jitter with a probability of jprob */
  if (random() <= jprob) {
    /* jitter up to jmax samples */
    joffset = random();
  }
#endif

#ifdef DEBUG_SHM
  fprintf(stderr, ", last_b = %p\n", *last_buffer);
#endif
  for (loop = lSector*CD_FRAMESAMPLES + joffset; 
       loop < (lSector+SectorBurstVal)*CD_FRAMESAMPLES + joffset; 
       loop++) {
    *q++ = 0;
    *q++ = 1;
  }
#ifdef DEBUG_SHM
  fprintf(stderr, "wrote from %p upto %p - 4, last_b = %p\n", p, q, *last_buffer);
#endif
  sim_pos = (lSector+SectorBurstVal)*CD_FRAMESAMPLES + joffset; 
}

static unsigned sim_indices;

/* read the table of contents (toc) via the ioctl interface */
static unsigned ReadToc_sim ( TOC *toc )
{
    unsigned int i;
    unsigned tracks;
    unsigned int scenario;
    int scen[12][3] = { 
      {1,1,5000}, 
      {1,2,5000}, 
      {1,99,150}, 
      {2,1,5000}, 
      {2,2,5000}, 
      {2,99,150},
      {5,1,5000}, 
      {5,2,5000}, 
      {5,99,150}, 
      {99,1,1000}, 
      {99,2,1000}, 
      {99,99,150}, 
    };

    fprintf(stderr, "select one of the following TOCs\n"
	    "0 :  1 track  with  1 index\n"
	    "1 :  1 track  with  2 indices\n"
	    "2 :  1 track  with 99 indices\n"
	    "3 :  2 tracks with  1 index each\n"
	    "4 :  2 tracks with  2 indices each\n"
	    "5 :  2 tracks with 99 indices each\n"
	    "6 :  5 tracks with  1 index each\n"
	    "7 :  5 tracks with  2 indices each\n"
	    "8 :  5 tracks with 99 indices each\n"
	    "9 : 99 tracks with  1 index each\n"
	    "10: 99 tracks with  2 indices each\n"
	    "11: 99 tracks with 99 indices each\n"
	    );

    do {
      scanf("%u", &scenario);
    } while (scenario > sizeof(scen)/2/sizeof(int));

    /* build table of contents */

    tracks = scen[scenario][0] + 1;
    sim_indices = scen[scenario][1];

    for (i = 0; i < tracks; i++) {
        toc[i].bFlags = 0x1b;
        toc[i].bTrack = i + 1;
        toc[i].dwStartSector = i * scen[scenario][2];
    }
    toc[tracks-1].bTrack = 0xaa;
    return --tracks;           /* without lead-out */
}


/* request sub-q-channel information. This function may cause confusion
 * for a drive, when called in the sampling process.
 */
static subq_chnl *ReadSubQ_sim ( unsigned char format, unsigned char track )
{
    subq_chnl *SQp = (subq_chnl *) (SubQbuffer);
    subq_position *SQPp = (subq_position *) &SQp->data;
    unsigned long sim_pos1;
    unsigned long sim_pos2;

    if ( format != GET_POSITIONDATA ) return NULL;  /* not supported by sim */

    /* simulate CDROMSUBCHNL ioctl */

    /* copy to SubQbuffer */
    SQp->audio_status 	= 0;
    SQp->format 	= 0xff;
    SQp->control_adr	= 0xff;
    sim_pos1 = sim_pos/CD_FRAMESAMPLES;
    sim_pos2 = sim_pos1 % 150;
    SQp->track 		= (sim_pos1 / 150) + 1;
    SQp->index 		= min(sim_pos2, sim_indices);
    SQPp->abs_min 	= sim_pos1 / (75*60);
    SQPp->abs_sec 	= (sim_pos1 / 75) % 60;
    SQPp->abs_frame 	= sim_pos1 % 75;
    SQPp->trel_min 	= sim_pos2 / (75*60);
    SQPp->trel_sec 	= (sim_pos2 / 75) % 60;
    SQPp->trel_frame 	= sim_pos2 % 75;

    return (subq_chnl *)(SubQbuffer);
}

static void SetupSimCd(void)
{
    EnableCdda =  (void (*)(int))Dummy;
    ReadCdRom = ReadCdRom_sim;
    ReadToc = ReadToc_sim;
    ReadSubQ = ReadSubQ_sim;

#if 0
    fprintf( stderr, "Please enter values for 'global.littleendian',  'overlap' :");
    scanf("%d,%d", &global.littleendian, &global.overlap); 
#else
    global.littleendian = 1;
    global.overlap = min(global.nsectors-1, 1);
#endif
}
#endif /* def SIM_CD */

/* perform initialization depending on the interface used. */
void SetupInterface( void )
{
    /* build signal set to block for during generic scsi */
    sigemptyset (&signalset);
    sigaddset (&signalset, SIGINT);
    sigaddset (&signalset, SIGPIPE);
    sigac.sa_handler = exit;
    sigemptyset(&sigac.sa_mask);
    sigac.sa_flags = 0;
    sigaction( SIGINT, &sigac, NULL);
    sigaction( SIGQUIT, &sigac, NULL);
    sigaction( SIGTERM, &sigac, NULL);
    sigaction( SIGHUP, &sigac, NULL);
    sigaction( SIGPIPE, &sigac, NULL);
    sigaction( SIGTRAP, &sigac, NULL);
    sigaction( SIGIOT, &sigac, NULL);

#if	defined SIM_CD
    global.nsectors = 75;
    fprintf( stderr, "SIMULATION MODE !!!!!!!!!!!\n");
#else
    /* ensure interface is setup correctly */
    global.cd_fd = OpenCdRom ( global.dev_name );

#if defined(HPUX)
	
	/* gain exclusive access to the CDROM-drive */
	if (ioctl(global.cd_fd, SIOC_EXCLUSIVE, 1) < 0) {
		perror("could not get exclusive access to the specified device");
		exit(-1);
	}


#else
#if defined(_AIX)


#else
#endif
#endif
#endif

    /* Value of 'nsectors' must be defined here */
    assert(global.nsectors > 0);

    fill_buffer = request_shm_sem(2*sizeof(*fill_buffer) + 
                    global.buffers*(entry_size),
                    (unsigned char **)&fill_buffer);
    last_buffer = (myringbuff **)((char *)fill_buffer + sizeof(*fill_buffer));
    bufferCdda = (unsigned char *)(BASE->data);
    set_total_buffers(global.buffers, sem_id);

    /* request one sector for table of contents */
    bufferTOC = malloc( CD_FRAMESIZE );      /* assumes sufficient aligned addresses */
    /* SubQchannel buffer */
    SubQbuffer = malloc( 48 );               /* assumes sufficient aligned addresses */
    cmd = malloc( 18 );                      /* assumes sufficient aligned addresses */
    if ( !bufferTOC || !SubQbuffer || !cmd ) {
       fprintf( stderr, "Too low on memory. Giving up.\n");
       exit(2);
    }

#if	defined SIM_CD
    SetupSimCd();
#else
    adjust_ssize = 1;
    /* if drive is of type scsi, get vendor name */
    if (interface == GENERIC_SCSI) {
        unsigned sector_size;

	SetupSCSI();
        sector_size = get_orig_sectorsize(&orgmode4, &orgmode10, &orgmode11);
        if ( sector_size != 2048 && set_sectorsize(2048) ) {
	  fprintf( stderr, "Could not change sector size from %d to 2048\n", sector_size );
	  adjust_ssize = 2048 / sector_size;
        }
#if defined (__linux__)
    } else {
	SetupCookedIoctl( global.dev_name );
#endif
    }
#endif	/* if def SIM_CD */
}
